import io
import os
import re
from urllib.parse import urlparse

import paramiko

from redis_file_transfer.filter import filename_filter
from redis_file_transfer.send import Sender


class Fetcher:
    delete = False
    include = ""
    exclude = "$^"
    channel = "default"

    def __init__(self, redis_url, fetch_url) -> None:
        self._redis_url = redis_url
        self._url = fetch_url
        parse_object = urlparse(self._url)
        self._hostname = parse_object.hostname
        self._port = parse_object.port
        self._username = parse_object.username
        self._password = parse_object.password
        self._path = parse_object.path
        if not self._port:
            self._port = 22

    def fetch(self):
        tp = paramiko.Transport((self._hostname, self._port))
        tp.connect(
            username=self._username, password=self._password
        )  # , hostkey=hostFingerprint)
        sftpClient = paramiko.SFTPClient.from_transport(tp)
        if sftpClient:
            for file in sftpClient.listdir(self._path):
                if filename_filter(
                    file, re.compile(self.include), re.compile(self.exclude)
                ):
                    filepath = os.path.join(self._path, file)
                    with io.BytesIO() as fo:
                        sftpClient.getfo(filepath, fo)
                        print(f"File fetched, filename: {file}")
                        fo.seek(0)
                        data = fo.read()
                    send = Sender(self._redis_url, file)
                    send.channel = self.channel
                    file_data = send._generate_file_data(data)
                    if file_data:
                        send._send_file_data(file_data)
                    if self.delete:
                        sftpClient.remove(filepath)
            sftpClient.close()
        tp.close()



import argparse
import sys
from typing import Optional, Sequence

import redis

from redis_file_transfer.fetch import Fetcher
from redis_file_transfer.receive import Receiver
from redis_file_transfer.send import Sender


def _add_redis_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--redis-url",
        default="redis://localhost:6379",
        help="redis url, default=redis://localhost:6379",
    )


def _add_channel_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--channel",
        "-c",
        default="default",
        help="channel for transfers, default=default",
    )


def main(argv: Optional[Sequence[str]] = None) -> int:

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command", title="commands")
    subparsers.required = True

    send_parser = subparsers.add_parser("send", help="send files")
    send_parser.add_argument("filename")
    _add_redis_options(send_parser)
    _add_channel_options(send_parser)
    send_parser.add_argument(
        "--delete",
        "-d",
        action="store_true",
        help="delete file after successfully transfered",
    )

    receive_parser = subparsers.add_parser("receive", help="receive files")
    receive_parser.add_argument("directory", help="directory to store received files")
    _add_redis_options(receive_parser)
    _add_channel_options(receive_parser)
    receive_parser.add_argument(
        "--group",
        "-g",
        help="receiving group, generated id as default",
    )
    receive_parser.add_argument(
        "--overwrite",
        "-o",
        action="store_true",
        help="overwrite file or append 001,002..etc to filename if exists",
    )
    receive_parser.add_argument(
        "--include",
        default="",
        help="regex for file inclusion, default: %(default)r",
    )
    receive_parser.add_argument(
        "--exclude",
        default="$^",
        help="regex for file exclusion, default: %(default)r",
    )

    fetch_parser = subparsers.add_parser("fetch", help="fetch files from sftp")
    _add_redis_options(fetch_parser)
    fetch_parser.add_argument("url")
    _add_channel_options(fetch_parser)
    fetch_parser.add_argument(
        "--delete",
        "-d",
        action="store_true",
        help="delete source file when fetched",
    )
    fetch_parser.add_argument(
        "--include",
        default="",
        help="regex for file inclusion, default: %(default)r",
    )
    fetch_parser.add_argument(
        "--exclude",
        default="$^",
        help="regex for file exclusion, default: %(default)r",
    )

    args = parser.parse_args(argv)

    try:
        if args.command == "send":
            try:
                sender = Sender(redis_url=args.redis_url, filename=args.filename)
                if args.channel:
                    sender.channel = args.channel
                sender.send()
            except FileNotFoundError:
                print(f"Error: file {args.filename} not found", file=sys.stderr)
                return 2
        elif args.command == "receive":
            try:
                receiver = Receiver(redis_url=args.redis_url, directory=args.directory)
                receiver.overwrite = args.overwrite
                receiver.channel = args.channel
                if args.group:
                    receiver.group_name = args.group
                receiver.include = args.include
                receiver.exclude = args.exclude
                receiver.run()
            except FileNotFoundError:
                print(f"Error: directory {args.directory} not found", file=sys.stderr)
                return 2
        elif args.command == "fetch":
            fetcher = Fetcher(redis_url=args.redis_url, fetch_url=args.url)
            fetcher.channel = args.channel
            fetcher.delete = args.delete
            fetcher.include = args.include
            fetcher.exclude = args.exclude
            fetcher.fetch()
    except redis.exceptions.ConnectionError:
        print(f"Error: can't connect to redis, url: {args.redis_url}", file=sys.stderr)
        return 2

    return 0


if __name__ == "__main__":
    SystemExit(main())



import base64
import os
import re

import redis

from redis_file_transfer.client_id import generate_id
from redis_file_transfer.filter import filename_filter


class Receiver:
    channel: str = "default"
    group_name: str = ""
    overwrite: bool = False
    directory: str = ""
    last_message_id: str = ""
    exclude: str = "$^"
    include: str = ""

    def __init__(self, redis_url: str, directory: str):
        self._redis = redis.from_url(redis_url, encoding="utf8", decode_responses=True)
        self.directory = directory
        self._check_directory()

    def _check_directory(self):
        if not os.path.exists(self.directory):
            raise FileNotFoundError(f"Directory '{self.directory}' does not exists")

    def _group_exists(self):
        if self._redis.exists(self.channel):
            for group in self._redis.xinfo_groups(self.channel):
                if group["name"] == self.group_name:
                    return True
        return False

    def _create_group(self):
        self._redis.xgroup_create(
            self.channel, self.group_name, self.last_message_id, True
        )

    def run(self):
        if not self._group_exists:
            self._create_group()
        if not self.group_name:
            self.group_name = generate_id(
                self.channel, self.directory, self.include, self.exclude
            )
        self._get_pending_files()
        self._redis.close()

    def _get_pending_files(self):
        self.last_message_id = self._get_last_message_id()
        events = self._redis.xread({self.channel: self.last_message_id})
        if not events:
            print("No new files receive.")
            return False
        for _stream, rows in events:
            for message_id, row in rows:
                self._save_received_file(message_id, row)
                self._save_last_message_id(message_id)
        return True

    def _save_received_file(self, message_id: str, data: dict):
        if filename_filter(
            data["filename"], re.compile(self.include), re.compile(self.exclude)
        ):
            print(f"Received file: {data['filename']}, id: {message_id}")
            filename = os.path.join(self.directory, data["filename"])
            if not self.overwrite and os.path.exists(filename):
                filename = self._get_new_filename(filename)
            with open(filename, "wb") as file:
                file_data = base64.decodebytes(str(data["data"]).encode("utf8"))
                file.write(file_data)
        else:
            print(
                f"Skipped file: {data['filename']}, id: {message_id}, matched exclude filter: {self.exclude}"
            )

    def _save_last_message_id(self, message_id: str) -> None:
        key = self._key
        self._redis.set(key, message_id)

    @property
    def _key(self):
        return f"rft_{self.group_name}_last_id"

    def _get_last_message_id(self) -> str:
        key = self._key
        message_id = self._redis.get(key)
        if message_id:
            return message_id
        else:
            if self._redis.exists(self.channel):
                channel_data = self._redis.xinfo_stream(self.channel)
                if channel_data:
                    message_id = channel_data["last-generated-id"]
                    self._redis.set(key, message_id)
                    return message_id
            else:
                message_id = "0-0"
                self._redis.set(key, message_id)
                return message_id

    def _get_new_filename(self, filename):
        num = 1
        temp = os.path.splitext(filename)
        new_filename = f"{temp[0]}_{str(num).zfill(3)}{temp[1]}"
        while os.path.exists(new_filename):
            num += 1
            new_filename = f"{temp[0]}_{str(num).zfill(3)}{temp[1]}"
        return new_filename



import hashlib


def generate_id(channel: str, directory: str, include_filter: str, exclude_filter: str):
    data_str = f"{channel}|{directory}|{include_filter}|{exclude_filter}"
    md5 = hashlib.md5()
    md5.update(data_str.encode("utf8"))
    return md5.hexdigest()



import base64
import datetime
import os

import redis


class Sender:
    channel: str = "default"
    filename: str = ""

    def __init__(self, redis_url: str, filename: str):
        self._redis = redis.from_url(redis_url, encoding="utf8", decode_responses=True)
        self.filename = filename

    def _file_exists(self):
        if not os.path.exists(self.filename):
            raise FileNotFoundError(f"File '{self.filename}' not found")
        return True

    def send(
        self,
    ):
        if self._file_exists():
            file_data = self._load_file()
            if file_data:
                self._send_file_data(file_data)

    def _load_file(self):
        filedata = None
        with open(self.filename, "rb") as file:
            file_data = file.read()
        filedata = self._generate_file_data(file_data)
        return filedata

    def _generate_file_data(self, filebytes: bytes):
        filedata = None
        data = base64.encodebytes(filebytes)
        if data:
            filedata = {
                "filename": self.filename,
                "data": data,
                "channel": self.channel,
                "filetime_created": str(datetime.datetime.now()),
            }
        return filedata

    def _clean_old_done_files(self):
        lowest_id = ""
        if self._redis.exists(self.channel):
            groups_data = self._redis.xinfo_groups(self.channel)
            for group in groups_data:
                last_id = self._redis.get(f"rft_{group['name']}_last_id")
                if not lowest_id:
                    lowest_id = last_id
                if last_id and last_id != "0-0" and last_id < lowest_id:
                    lowest_id = last_id
        if lowest_id:
            self._redis.execute_command("XTRIM", self.channel, "MINID", lowest_id)

    def _send_file_data(
        self,
        file_data: dict,
    ):
        result = self._redis.xadd(
            self.channel,
            file_data,
        )
        print(f"File sent, filename: {self.filename}, id: {result}")
        self._clean_old_done_files()
        self._redis.close()
